/*
 * List the contents of an ISIS-II disk image or extract the files.
 *
 * Written by the OS/2 Museum, 2014.
 * This source code is hereby placed in the public domain.
 */

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <assert.h>

/* The ISIS-II floppies use double-sided media with 26 sectors per track
 * and 77 tracks. Sector size is 128 bytes, for a total capacity of
 * 77 * 26 * 2 * 128 = 512,512 bytes (500.5 KB). Logical addressing only
 * uses one side with 52 sectors.
 */

/* ISIS-II sector size. */
#define SECT_SZ             128

/* ISIS-II sectors per track. */
#define I2D_SPT             (26 * 2)

/* ISIS-II disk size in bytes. */
#define I2D_DSK_SZ          (77 * I2D_SPT * SECT_SZ)

typedef union i2d_blkadr_u {
    uint16_t        addr;           /* 16-bit block address. */
    struct {
        uint8_t     sector;         /* Sector number (1-based). */
        uint8_t     track;          /* Track number (0-based). */
    } s;
} i2d_blkadr;

/* ISIS-II Directory Entry. */
typedef struct i2d_dirent_s {
    uint8_t     stat;       /* Entry status. */
    char        name[6];    /* The 6.3 name/extension. */
    char        ext[3];
    uint8_t     attr;       /* Attribute byte. */
    uint8_t     last_sz;    /* Number of bytes used in last block. */
    uint16_t    num_blks;   /* Number of blocks occupied by file. */
    i2d_blkadr  first_lblk; /* Address of the first linkage block. */
} i2d_dirent;

/* Values for the directory entry status byte. */
#define I2D_DS_VALID        0x00    /* Entry/file in use. */
#define I2D_DS_UNUSED       0x7F    /* Not yet used entry. */
#define I2D_DS_DELETED      0xFF    /* Deleted file. */

/* Attribute bit values. */
#define I2D_ATTR_HID        0x01    /* Hidden (invisible) file. */
#define I2D_ATTR_SYS        0x02    /* System file. */
#define I2D_ATTR_RDO        0x04    /* Read-only (write protected) file. */
#define I2D_ATTR_FMT        0x80    /* Format file (not user data). */

/* ISIS-II Linkage Block. */
typedef struct i2d_lnkblk_s {
    i2d_blkadr  prev_blk;   /* Previous linkage block (0 if none). */
    i2d_blkadr  next_blk;   /* Next linkage block (0 if none). */
    i2d_blkadr  blks[62];   /* Blocks occupied by file. */
} i2d_lnkblk;

#define I2D_DBLK_PER_LBLK   62      /* Data blocks per linkage block. */


typedef FILE *img_handle;


int img_read_sect( img_handle imgf, unsigned sect_no, void *dest )
{
    fseek( imgf, sect_no * SECT_SZ, SEEK_SET );
    return( fread( dest, 1, SECT_SZ, imgf ) );
}


long img_query_size( img_handle imgf )
{
    fseek( imgf, 0, SEEK_END );
    return( ftell( imgf ) );
}


int dsk_read_sector( img_handle imgf, i2d_blkadr blk_adr, void *dest )
{
    int         rc;
    unsigned    sect_no;

    assert( blk_adr.s.sector > 0 );
    sect_no = blk_adr.s.track * I2D_SPT + blk_adr.s.sector - 1;
    rc = img_read_sect( imgf, sect_no, dest );
    return( rc );
}


int file_read_blk( img_handle imgf, i2d_blkadr anchor, unsigned blk_no, void *dest )
{
    i2d_blkadr      lblk_adr = anchor;
    i2d_lnkblk      link_blk;
    int             rc;

    /* Go through the linkage blocks until we're within 62 blocks. */
    rc = dsk_read_sector( imgf, lblk_adr, &link_blk );
    while( blk_no >= I2D_DBLK_PER_LBLK ) {
        lblk_adr = link_blk.next_blk;
        blk_no  -= I2D_DBLK_PER_LBLK;
        rc = dsk_read_sector( imgf, lblk_adr, &link_blk );
    };

    /* Read the actual data block. */
    rc = dsk_read_sector( imgf, link_blk.blks[blk_no], dest );

    return( rc );
}


int read_dir( img_handle imgf, i2d_dirent *dir, unsigned n_ent )
{
    i2d_blkadr      dir_link;
    unsigned        i;
    int             rc = 0;

    /* Directory linkage block is always at 1, 1 (i.e. the first sector
     * on the second track).
     */
    dir_link.s.track  = 1;
    dir_link.s.sector = 1;

    for( i = 0; i < n_ent * sizeof( i2d_dirent ) / SECT_SZ; ++i ) {
        rc = file_read_blk( imgf, dir_link, i, &dir[i * SECT_SZ / sizeof( i2d_dirent ) ] );
    }
    return( rc );
}


int extract_file( img_handle imgf, i2d_dirent *dir )
{
    FILE        *of;
    char        fname[6 + 1 + 3 + 1];
    uint8_t     buf[SECT_SZ];
    unsigned    nbytes;
    unsigned    blk = 0;
    int         rc  = 0;

    /* Form the file name based on whether an extension is present. */
    if( dir->ext[0] )
        sprintf( fname, "%.6s.%.3s", dir->name, dir->ext );
    else
        sprintf( fname, "%.6s", dir->name );

    printf( "Extracting %s\n", fname );

    /* Create the file. */
    of = fopen( fname, "wb" );
    if( !of ) {
        printf( "Failed to create `%s'\n", fname );
        rc = 1;
    }

    /* Copy file blocks (if any). */
    while( blk < dir->num_blks ) {
        rc = file_read_blk( imgf, dir->first_lblk, blk, buf );
        /* The last block may be smaller. */
        if( blk == dir->num_blks - 1 )
            nbytes = dir->last_sz;
        else
            nbytes = SECT_SZ;
        rc = fwrite( buf, 1, nbytes, of );
        ++blk;
    }

    fclose( of );

    return( rc );
}


int process_image( img_handle imgf, int do_ext )
{
    i2d_dirent  dir[128];
    unsigned    i;
    unsigned    fsize;
    int         rc = 0;

    /* Read the directory blocks and run through the entries. */
    rc = read_dir( imgf, dir, sizeof( dir ) / sizeof( dir[0] ) );

    for( i = 0; i < sizeof( dir ) / sizeof( dir[0] ); ++i ) {
        if( dir[i].stat == I2D_DS_VALID ) {
            /* Calculate the file size. Assume zero-size files are valid. */
            assert( dir[i].last_sz <= SECT_SZ );
            if( dir[i].num_blks )
                fsize = dir[i].num_blks * SECT_SZ - SECT_SZ + dir[i].last_sz;
            else
                fsize = 0;
            if( do_ext ) {
                rc = extract_file( imgf, &dir[i] );
            }
            else
                printf( "%-6.6s %-3.3s  %6u bytes\n", dir[i].name, dir[i].ext, fsize );
        }
    }
    return( rc );
}


int main( int argc, char **argv )
{
    FILE    *f;
    int     rc;
    int     f_arg;
    int     do_extract = 0;
    long    img_size;

    f_arg = 1;
    if( argc > 1 && !strcmp( argv[1], "-e" ) ) {
        do_extract = 1;
        ++f_arg;
    }

    if( argc <= f_arg ) {
        printf( "usage: %s [-e] <imgfile>\n", argv[0] );
        return( 1 );
    }

    f = fopen( argv[f_arg], "rb" );
    if( !f ) {
        printf( "Failed to open `%s' for reading!\n", argv[f_arg] );
        return( 2 );
    }

    img_size = img_query_size( f );
    if( img_size == I2D_DSK_SZ ) {
        rc = process_image( f, do_extract );
    } else {
        printf( "Incorrect image size!\n" );
        rc = 3;
    }

    fclose( f );
    return( rc );
}
